
# Import opinel
from opinel.utils import *

import time

########################################
##### STS-related arguments
########################################

#
# Add an STS-related argument to a recipe
#
def add_sts_argument(parser, argument_name):
    if argument_name == 'mfa-serial':
        parser.add_argument('--mfa-serial',
                            dest='mfa_serial',
                            default=[ None ],
                            help='MFA device\'s serial number')
    elif argument_name == 'mfa-code':
        parser.add_argument('--mfa-code',
                            dest='mfa_code',
                            default=[ None ],
                            help='MFA code')
    elif argument_name == 'external-id':
        parser.add_argument('--external-id',
                            dest='external_id',
                            default=None,
                            help='External ID')
    else:
        raise Exception('Invalid parameter name: %s' % argument_name)


########################################
##### Helpers
########################################

#
# Get role credentials
#
def assume_role(sts_client, profile_name, role_arn, role_session_name, mfa_serial = None, mfa_code = None, external_id = None):
    sts_args = {
      'RoleArn': role_arn,
      'RoleSessionName': role_session_name
    }

    if mfa_serial and mfa_code:
      sts_args['TokenCode'] = mfa_code
      sts_args['SerialNumber'] = mfa_serial

    if external_id:
      sts_args['ExternalId'] = external_id

    sts_response = sts_client.assume_role(**sts_args)
    with open('/home/loic/.aws/cli/cache/%s-%s.json' % (profile_name, role_arn.replace(':', '_')), 'wt+') as f:
        f.write(sts_response)
    return sts_response['Credentials']['AccessKeyId'], sts_response['Credentials']['SecretAccessKey'], sts_response['Credentials']['SessionToken']

#
# Fetch role credentials and store them in a file
#
def assume_role_and_save_in_credentials(source_profile, profile_name, role_arn, mfa_serial_arg, mfa_code, external_id = None):

    # Init
    key_id = secret = mfa_serial = session_token = None
    save_no_mfa_credentials = False

    # Search credentials
    if mfa_code:
        key_id, secret, mfa_serial, session_token = read_creds_from_aws_credentials_file('%s-nomfa' % source_profile)
    else:
        key_id, secret, mfa_serial, session_token = read_creds_from_aws_credentials_file(source_profile)

    # If nothing found, search for long-lived credentials stored as -nomfa
    if (not key_id or not secret) and not profile_name.endswith('-nomfa'):
        key_id, secret, mfa_serial, session_token = read_creds_from_aws_credentials_file(profile_name + '-nomfa')
        save_no_mfa_credentials = True

    if not key_id or not secret:
        raise Exception('Error: could not find AWS credentials. Use the --help option for more information.')

    # If an MFA serial was provided as an argument, discard whatever we found in config file
    if mfa_serial_arg:
        mfa_serial = mfa_serial_arg

    # Prompt for a token code if needed
    if mfa_serial and not mfa_code and not session_token:
        mfa_code = prompt_4_mfa_code()

    sts_client = connect_sts(key_id, secret, session_token)
    session_name = 'AWS-Recipes-%s' % int(time.time())
    role_key_id, role_secret, role_session_token = assume_role(sts_client, profile_name, role_arn, session_name, mfa_serial, mfa_code, external_id)

    # Save session credentials
#    role_profile = profile_name.replace('-nomfa', '') + '-' + role_session_name
#    write_creds_to_aws_credentials_file(role_profile, key_id = role_key_id, secret = role_secret, session_token = role_session_token)

    printInfo('Your role credentials have been cached to XXX.')

#
# Connect to STS
#
def connect_sts(key_id, secret, session_token):
    try:
        printInfo('Connecting to AWS STS...')
        return boto3.client('sts', aws_access_key_id = key_id, aws_secret_access_key = secret, aws_session_token = session_token)
    except Exception as e:
        printError('Error: could not connect to STS.')
        printException(e)
        return None

#
# Fetch STS credentials and store them in a file
#
def init_sts_session_and_save_in_credentials(profile_name, credentials_file = aws_credentials_file, mfa_code = None, mfa_serial_arg = None):

    # Init
    key_id = secret = mfa_serial = session_token = None
    save_no_mfa_credentials = False

    # First check for long-lived credentials stored under <profile_name>-nomfa
    if not mfa_serial_arg and not profile_name.endswith('-nomfa'):
        key_id, secret, mfa_serial, session_token = read_creds_from_aws_credentials_file(profile_name + '-nomfa')

    # If nothing found or the mfa serial was provided, search with the provided profile name
    if not key_id or not secret:
        key_id, secret, mfa_serial, session_token = read_creds_from_aws_credentials_file(profile_name)
        save_no_mfa_credentials = True

    if not key_id or not secret:
        raise Exception('Error: could not find AWS credentials. Use the --help option for more information.')

    if mfa_serial == mfa_serial_arg == None:
        raise Exception('Error: you must provide an MFA device serial number to receive STS credentials. Update your credentials file or use the --serial_number argument.')

    # If an MFA serial was provided as an argument, discard whatever we found in config file
    if mfa_serial_arg:
        mfa_serial = mfa_serial_arg

    # Get STS credentials
    session_access_key, session_secret_key, session_token = init_sts_session(key_id, secret, mfa_serial, mfa_code)

    # Save long-lived credentials if needed
    if save_no_mfa_credentials:
        write_creds_to_aws_credentials_file(profile_name + '-nomfa', key_id = key_id, secret = secret, mfa_serial = mfa_serial)

    # Save session credentials
    write_creds_to_aws_credentials_file(profile_name, key_id = session_access_key, secret = session_secret_key, session_token = session_token)

    # Success
    printInfo('Successfully configured the session token for profile \'%s\'.' % profile_name)




# Import opinel
from opinel.utils import *


########################################
##### Helpers
########################################

#
# Connect to RDS
#
def connect_rds(key_id, secret, session_token, region_name):
    return connect_service('rds', key_id, secret, session_token, region_name)




# Import opinel
from opinel.utils import *


########################################
##### Helpers
########################################

#
# Connect to Redshift
#
def connect_redshift(key_id, secret, session_token, region):
    return connect_service('redshift', key_id, secret, session_token, region)




# Import opinel
from opinel.utils import *


########################################
##### Helpers
########################################

#
# Connect to EC2 API
#
def connect_ec2(key_id, secret, session_token,  region_name):
    return connect_service('ec2', key_id, secret, session_token, region_name)

#
# Connect to ELB API
#
def connect_elb(key_id, secret, session_token, region_name):
    return connect_service('elb', key_id, secret, session_token, region_name)


########################################
##### Helpers
########################################

#
# Get name from tags
#
def get_name(local, remote, default_attribute):
    name_found = False
    if 'Tags' in remote:
        for tag in remote['Tags']:
            if tag['Key'] == 'Name':
                local['name'] = tag['Value']
                name_found = True
    if not name_found:
        local['name'] = remote[default_attribute]



from __future__ import print_function

from opinel.load_data import load_data

import json
import os
import sys

#master_permissions_path = os.path.join(
#    os.path.dirname(os.path.realpath(__file__)),
#    'master_permissions.json')

# To update the master_permissions.json, open this URL in chrome:
# https://awspolicygen.s3.amazonaws.com/policygen.html
# Bring up the console from the inspector and type:
# JSON.stringify(app.PolicyEditorConfig.serviceMap)
# Copy and Paste the output into a file and then reformat that JSON with this code:

# with open(".../new.json") as json_data:
#     blah = json.load(json_data)
#
# with open('.../master_permissions.json', 'w') as outfile:
#     json.dump(blah, outfile, indent=2, sort_keys=True)

# Make sure indent=2 and sort_keys=True.


#with open(master_permissions_path) as json_data:
#    global_permissions = json.load(json_data)
#    json_data.close()
global_permissions = load_data('master_permissions.json')

all_permissions = set()

for technology_name in global_permissions:
    technology_prefix = global_permissions[technology_name]["StringPrefix"]
    for action in global_permissions[technology_name]["Actions"]:
        all_permissions.add("{}:{}".format(technology_prefix, action.lower()))


policy_headers = ['rolepolicies', 'grouppolicies', 'userpolicies', 'policy']


def expand_minimize_over_policies(policies, activity, **kwargs):
    for header in policy_headers:
        if header in policies:
            output = {header: {}}
            for policy in policies[header]:
                output[header][policy] = activity(policy=policies[header][policy], **kwargs)
            return output

    return activity(policy=policies, **kwargs)


def _get_prefixes_for_action(action):
    """
    :param action: iam:cat
    :return: [ "iam:", "iam:c", "iam:ca", "iam:cat" ]
    """
    (technology, permission) = action.split(':')
    retval = ["{}:".format(technology)]
    phrase = ""
    for char in permission:
        newphrase = "{}{}".format(phrase, char)
        retval.append("{}:{}".format(technology,newphrase))
        phrase = newphrase
    return retval


def _expand_wildcard_action(action):
    """
    :param action: 'autoscaling:*'
    :return: A list of all autoscaling permissions matching the wildcard
    """
    if isinstance(action, list):
        expanded_actions = []
        for item in action:
            expanded_actions.extend(_expand_wildcard_action(item))
        return expanded_actions

    else:
        if '*' in action:
            pre_wildcard = action.split('*')[0]
            expanded = [expanded_action.lower() for expanded_action in all_permissions if expanded_action.startswith(pre_wildcard.lower())]

            # if we get a wildcard for a tech we've never heard of, just return the wildcard
            if not expanded:
                return [action.lower()]

            return expanded

        return [action.lower()]

    raise Exception("Action must be a list or a string")


def _get_desired_actions_from_statement(statement):
    desired_actions = set()
    actions = _expand_wildcard_action(statement['Action'])

    for action in actions:
        if action not in all_permissions:
            raise Exception("Desired action not found in master permission list. {}".format(action))
        desired_actions.add(action)

    return desired_actions


def _get_denied_prefixes_from_desired(desired_actions):
    denied_actions = all_permissions.difference(desired_actions)
    denied_prefixes = set()
    for denied_action in denied_actions:
        for denied_prefix in _get_prefixes_for_action(denied_action):
            denied_prefixes.add(denied_prefix)

    return denied_prefixes


def _check_min_permission_length(permission, minchars=None):
    if minchars and len(permission) < int(minchars) and permission != '':
        print("Skipping prefix {} because length of {}".format(permission, len(permission)) , file = sys.stderr)
        return True
    return False


def minimize_statement_actions(statement, minchars=None):
    minimized_actions = set()

    if statement['Effect'] != 'Allow':
        raise Exception("Minification does not currently work on Deny statements.")

    desired_actions = _get_desired_actions_from_statement(statement)
    denied_prefixes = _get_denied_prefixes_from_desired(desired_actions)

    for action in desired_actions:
        if action in denied_prefixes:
            print("Action is a denied prefix. Action: {}".format(action))
            minimized_actions.add(action)
            continue

        found_prefix = False
        prefixes = _get_prefixes_for_action(action)
        for prefix in prefixes:

            permission = prefix.split(':')[1]
            if _check_min_permission_length(permission, minchars=minchars):
                continue

            if prefix not in denied_prefixes:
                if prefix not in desired_actions:
                    prefix = "{}*".format(prefix)
                minimized_actions.add(prefix)
                found_prefix = True
                break

        if not found_prefix:
            print("Could not suitable prefix. Defaulting to {}".format(prefixes[-1]))
            minimized_actions.add(prefixes[-1])

    # sort the actions
    minimized_actions_list = list(minimized_actions)
    minimized_actions_list.sort()

    return minimized_actions_list


def get_actions_from_statement(statement):
    allowed_actions = set()

    if not type(statement.get('Action', [])) == list:
        statement['Action'] = [statement['Action']]

    for action in statement.get('Action', []):
        allowed_actions = allowed_actions.union(set(_expand_wildcard_action(action)))

    if not type(statement.get('NotAction', [])) == list:
        statement['NotAction'] = [statement['NotAction']]

    inverted_actions = set()
    for action in statement.get('NotAction', []):
        inverted_actions = inverted_actions.union(set(_expand_wildcard_action(action)))

    if inverted_actions:
        actions = _invert_actions(inverted_actions)
        allowed_actions = allowed_actions.union(actions)

    return allowed_actions


def _invert_actions(actions):
#    from iampoliciesgonewild import all_permissions
    return all_permissions.difference(actions)


def expand_policy(policy=None, expand_deny=False):

    # str_pol = json.dumps(policy, indent=2)
    # size = len(str_pol)

    if type(policy['Statement']) is dict:
        policy['Statement'] = [policy['Statement']]
    for statement in policy['Statement']:
        if statement['Effect'].lower() == 'deny' and not expand_deny:
            continue
        actions = get_actions_from_statement(statement)
        if 'NotAction' in statement:
            del statement['NotAction']
        statement['Action'] = sorted(list(actions))

    # str_end_pol = json.dumps(policy, indent=2)
    # end_size = len(str_end_pol)

    # print str_end_pol
    # print >> sys.stderr, "Start size: {}. End size: {}".format(size, end_size)
    return policy


def minimize_policy(policy=None, minchars=None):

    str_pol = json.dumps(policy, indent=2)
    size = len(str_pol)

    for statement in policy['Statement']:
        minimized_actions = minimize_statement_actions(statement, minchars=minchars)
        statement['Action'] = minimized_actions

    str_end_pol = json.dumps(policy, indent=2)
    end_size = len(str_end_pol)

    # print str_end_pol
    print("Start size: {}. End size: {}".format(size, end_size), file = sys.stderr)
    return policy
